/*
 * Remmina - The GTK+ Remote Desktop Client
 * Copyright (C) 2010 Vic Lee 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA.
 */

#ifndef __REMMINAKEYCHOOSER_H__
#define __REMMINAKEYCHOOSER_H__

G_BEGIN_DECLS

#define REMMINA_TYPE_KEY_CHOOSER               (remmina_key_chooser_get_type ())
#define REMMINA_KEY_CHOOSER(obj)               (G_TYPE_CHECK_INSTANCE_CAST ((obj), REMMINA_TYPE_KEY_CHOOSER, RemminaKeyChooser))
#define REMMINA_KEY_CHOOSER_CLASS(klass)       (G_TYPE_CHECK_CLASS_CAST ((klass), REMMINA_TYPE_KEY_CHOOSER, RemminaKeyChooserClass))
#define REMMINA_IS_KEY_CHOOSER(obj)            (G_TYPE_CHECK_INSTANCE_TYPE ((obj), REMMINA_TYPE_KEY_CHOOSER))
#define REMMINA_IS_KEY_CHOOSER_CLASS(klass)    (G_TYPE_CHECK_CLASS_TYPE ((klass), REMMINA_TYPE_KEY_CHOOSER))
#define REMMINA_KEY_CHOOSER_GET_CLASS(obj)     (G_TYPE_INSTANCE_GET_CLASS ((obj), REMMINA_TYPE_KEY_CHOOSER, RemminaKeyChooserClass))

typedef struct _RemminaKeyChooser
{
	GtkButton button;

	guint keyval;
} RemminaKeyChooser;

typedef struct _RemminaKeyChooserClass
{
	GtkButtonClass parent_class;
} RemminaKeyChooserClass;

GType remmina_key_chooser_get_type(void)
G_GNUC_CONST;

GtkWidget* remmina_key_chooser_new(guint keyval);
void remmina_key_chooser_set_keyval(RemminaKeyChooser *kc, guint keyval);

G_END_DECLS

#endif  /* __REMMINAKEYCHOOSER_H__  */

